SystemOrganization addCategory: #ConnectorsGraphLayout!

TestCase subclass: #NCGraphLayoutAlgorithmTest
	instanceVariableNames: 'mainWindowMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ConnectorsGraphLayout'!

!NCGraphLayoutAlgorithmTest methodsFor: 'running' stamp: 'drm 7/10/2003 18:21'!
addRectangleNamed: name at: pos

| r |
r := RectangleMorph new 
	position: pos;
	color: Color veryLightGray.
r addMorphCentered: (TextMorph new contents: name).
mainWindowMorph addMorph: r.
^r
	! !

!NCGraphLayoutAlgorithmTest methodsFor: 'running' stamp: 'nk 3/8/2004 00:04'!
connect: m1 to: m2 
	"the direction is because I expect children to point at thier parents--not 
	vice-versa "
	| conn |
	conn := NCAAConnectorMorph fromMorph: m2 toMorph: m1.
	conn makeForwardArrow.
	mainWindowMorph addMorph: conn! !

!NCGraphLayoutAlgorithmTest methodsFor: 'running' stamp: 'nk 7/17/2004 15:25'!
runGraph
	| ncla |
	ncla := NCSimulatedAnnealingGraphLayoutAlgorithm new.
	WorldState
		addDeferredUIMessage: [ncla layoutMorph: mainWindowMorph]
! !

!NCGraphLayoutAlgorithmTest methodsFor: 'running' stamp: 'nk 7/17/2004 15:25'!
setUp
	| b |
	mainWindowMorph := PasteUpMorph new.
	mainWindowMorph color: Color veryVeryLightGray;
		 position: 10 @ 10;
		 extent: 800 @ 800;
		 borderWidth: 2;
		 borderColor: Color black.
	b := SimpleButtonMorph new.
	b color: Color transparent;
		 borderColor: Color red;
		 position: 13 @ 777;
		 label: 'delete';
		 target: mainWindowMorph;
		 actionSelector: #delete;
		 setBalloonText: 'delete this morph'.
	mainWindowMorph addMorphBack: b.
	mainWindowMorph openInWorld
! !

!NCGraphLayoutAlgorithmTest methodsFor: 'running' stamp: 'drm 7/26/2003 13:37'!
testComplexTree
| r1 r2 r3 r4 r5 r6 r7 |
r1 := self addRectangleNamed: 'r1' at: 50@50.
r2 := self addRectangleNamed: 'r2' at: 150@150.
r3 := self addRectangleNamed: 'r3' at: 250@250.
r4 := self addRectangleNamed: 'r4' at: 350@350.
r5 := self addRectangleNamed: 'r5' at: 50@150.
r6 := self addRectangleNamed: 'r6' at: 150@250.
r7 := self addRectangleNamed: 'r7' at: 250@350.

self connect: r1 to: r2.
self connect: r3 to: r2.
self connect: r5 to: r3.
self connect: r6 to: r3.
self connect: r7 to: r6.
self connect: r4 to: r1.

self runGraph! !

!NCGraphLayoutAlgorithmTest methodsFor: 'running' stamp: 'drm 7/26/2003 13:37'!
testK7
| r1 r2 r3 r4 r5 r6 r7 a |
r1 := self addRectangleNamed: 'r1' at: 50@50.
r2 := self addRectangleNamed: 'r2' at: 150@150.
r3 := self addRectangleNamed: 'r3' at: 250@250.
r4 := self addRectangleNamed: 'r4' at: 350@350.
r5 := self addRectangleNamed: 'r5' at: 50@150.
r6 := self addRectangleNamed: 'r6' at: 150@250.
r7 := self addRectangleNamed: 'r7' at: 250@350.

a := { r1. r2. r3. r4. r5. r6. r7. }.

1 to: 7 do: [ :i |
			(i + 1) to: 7 do: [ :j | 
				self connect: (a at: i) to: (a at: j) ] ].

self runGraph! !

!NCGraphLayoutAlgorithmTest methodsFor: 'running' stamp: 'drm 7/26/2003 13:36'!
testOffPage
| r1 r2 r3 |
r1 := self addRectangleNamed: 'r1' at: 0@0.
r2 := self addRectangleNamed: 'r2' at: 15@15.
r3 := self addRectangleNamed: 'r3' at: 25@25.

self connect: r1 to: r2.
self connect: r3 to: r2.

self runGraph

	! !

!NCGraphLayoutAlgorithmTest methodsFor: 'running' stamp: 'nk 7/17/2004 16:21'!
testSimpleTree
	| r1 r2 r3 |
	
	r1 := self addRectangleNamed: 'r1' at: 150@250.
	r2 := self addRectangleNamed: 'r2' at: 250@50.
	r3 := self addRectangleNamed: 'r3' at: 350@250.

	self connect: r1 to: r2.
	self connect: r3 to: r2.

	self runGraph

! !

Object subclass: #NCGraphLayoutAlgorithm
	instanceVariableNames: 'nodes edges aesthetics forces maxIterations scale layoutMorph iterations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ConnectorsGraphLayout'!

!NCGraphLayoutAlgorithm commentStamp: '<historical>' prior: 0!
I am a force-directed algorithm for graph layout.

Given a set of nodes and edges (all Morphs), I move the nodes (and possibly bend the edges) to give the most pleasing result with respect to a set of aesthetics.

These aesthetics have weights.

Subclasses use different strategies to guide this process (for instance, by simulated annealing).

At each step of the algorithm, the morphs are actually moved. This gives visual feedback, but slows down the solution of the layout.

Each of the aesthetics methods answers an IdentityDictionary<Morph->Point>, which represents unweighted (Cartesian) force vectors to apply to the various nodes.

Instance variables:

nodes		<OrderedCollection<Morph>>
	which nodes to move
edges		<OrderedCollection<NCConnectorMorph>>
	the edges to consider
aesthetics	<OrderedCollection<Symbol->weight>>
	the order in which to call various evaluation methods, and the weight to apply to their output force vector
forces		<IdentityDictionary<Morph->Point>>
	the sum of the weighted force vectors, per node!

!NCGraphLayoutAlgorithm class methodsFor: 'instance creation' stamp: 'nk 10/21/2003 15:50'!
layoutMorph: aMorph
	^(self new) layoutMorph: aMorph; yourself! !

!NCGraphLayoutAlgorithm class methodsFor: 'instance creation' stamp: 'nk 10/21/2003 15:51'!
layoutWorld
	^self layoutMorph: self currentWorld! !

!NCGraphLayoutAlgorithm class methodsFor: 'instance creation' stamp: 'nk 3/14/2002 13:04'!
new
	^super new initialize! !

!NCGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 3/14/2002 11:18'!
addAesthetic: aSymbol withWeight: aNumber
	aesthetics add: aSymbol -> aNumber.! !

!NCGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 10/21/2003 15:09'!
addEdge: anEdge 
	self edges: (self edges copyWith: anEdge)! !

!NCGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 10/21/2003 15:14'!
addNode: aNode 
	self nodes: (self nodes copyWith: aNode).
	self forces: (self forces copyWith: 0.0 @ 0.0)! !

!NCGraphLayoutAlgorithm methodsFor: 'aesthetics' stamp: 'nk 10/21/2003 14:57'!
aeCentripetal
	"I am an aesthetic that causes nodes to fly outwards from the centroid.
	My magnitude is 1/d^2, where d is the magnitude of the distance
	between nodes and the centroid.
	Since I consider all nodes, my time complexity is O(|V|)"

	| centroid centers retval |
	retval := OrderedCollection new.
	centers := self nodes collect: [:ea | ea center asFloatPoint].
	centroid := centers average.
	centers do: 
			[:center | 
			| delta mag2 |
			delta := center - centroid.
			delta := delta / scale.
			mag2 := delta x * delta x + (delta y * delta y) max: 1.0e-8.
			retval := retval , { 
								1.0 / mag2 * delta}].
	^retval! !

!NCGraphLayoutAlgorithm methodsFor: 'aesthetics' stamp: 'nk 8/16/2004 13:58'!
aeEdgeFlowHang
	"I am an aesthetic that causes edges to tend to hang down so their start nodes are above their end nodes. Since I consider all edges, my time complexity is O(|E|)"

	| retval upForce |
	retval := IdentityDictionary 
				newFrom: (self nodes collect: [:node | node -> (0 @ 0)]).
	upForce := 0 @ -0.01 * scale.
	self edges do: 
			[:edge | 
			| fromNode toNode |
			fromNode := edge startConnection.
			toNode := edge endConnection.
			toNode notNil & fromNode notNil 
				ifTrue: 
					[retval at: fromNode put: (retval at: fromNode) + upForce.
					retval at: toNode put: (retval at: toNode ifAbsentPut: [ 0@0 ]) - upForce]].
	^self nodes collect: [:ea | retval at: ea]! !

!NCGraphLayoutAlgorithm methodsFor: 'aesthetics' stamp: 'nk 10/21/2003 15:28'!
aeEdgeRepulsion
	"I am an aesthetic that causes nodes to stay away from the edges. 
	My magnitude is 1/d^2, where d is the magnitude of the distance 
	between nodes and the edge. 
	Since I consider all nodes, my time complexity is O(|V|)"
	| centers retval |
	retval := OrderedCollection new.
	centers := self nodes
				collect: [:ea | ea center asFloatPoint].
	centers
		do: [:center | 
			| delta mag2x mag2y | 
			delta := center x < self layoutMorph center x
						ifTrue: [center x - self layoutMorph left]
						ifFalse: [center x - self layoutMorph right].
			delta := delta / scale.
			mag2x := delta isZero
						ifTrue: [1.0e8]
						ifFalse: [delta sign / (delta * delta) max: 1.0e-8].
			delta := center y < self layoutMorph center y
						ifTrue: [center y - self layoutMorph top]
						ifFalse: [center y - self layoutMorph bottom].
			delta := delta / scale.
			mag2y := delta isZero
						ifTrue: [1.0e8]
						ifFalse: [delta sign / (delta * delta) max: 1.0e-8].
			retval addLast: mag2x @ mag2y].
	^ retval! !

!NCGraphLayoutAlgorithm methodsFor: 'aesthetics' stamp: 'nk 10/21/2003 14:57'!
aeMinEdgeLength
	"I am an aesthetic that causes edges to be shortened. My magnitude is d^2, where d is the magnitude of the distance between nodes. Since I consider all edges, my time complexity is O(|E|)"

	| retval |
	retval := IdentityDictionary 
				newFrom: (self nodes collect: [:node | node -> (0 @ 0)]).
	self edges do: 
			[:edge | 
			| fromNode toNode delta mag force |
			fromNode := edge startConnection.
			toNode := edge endConnection.
			delta := toNode center - fromNode center.
			delta := delta / scale.
			mag := delta r.
			force := mag * delta.
			retval at: fromNode put: (retval at: fromNode) + force.
			retval at: toNode put: (retval at: toNode) - force].
	^self nodes collect: [:ea | retval at: ea]! !

!NCGraphLayoutAlgorithm methodsFor: 'aesthetics' stamp: 'nk 10/21/2003 14:57'!
aeNodeRepulsion
	"I am an aesthetic that attempts to keep nodes apart. My magnitude is 1/d^2, where d is the magnitude of the distance between nodes. Since I consider all node combinations, my time complexity is O(|V|^2)"

	| retval |
	retval := Array new: self nodes size withAll: 0.0 @ 0.0.
	(1 to: self nodes size) combinations: 2
		atATimeDo: 
			[:arr | 
			| ai1 ai2 node otherNode delta mag2 force |
			node := self nodes at: (ai1 := arr at: 1).
			otherNode := self nodes at: (ai2 := arr at: 2).
			delta := (node center - otherNode center) asFloatPoint.
			delta := delta / scale.
			mag2 := delta x * delta x + (delta y * delta y) max: 1.0e-8.
			force := 1.0 / mag2 * delta.
			retval at: ai1 put: (retval at: ai1) + force.
			retval at: ai2 put: (retval at: ai2) - force].
	^retval! !

!NCGraphLayoutAlgorithm methodsFor: 'aesthetics' stamp: 'nk 10/21/2003 14:57'!
aeSpringEdgeLength
	"I am an aesthetic that causes edges to tend to a specific length, resisting compression and extension. My magnitude is d, where d is the difference between an edges length and it's desired length. Since I consider all edges, my time complexity is O(|E|)"

	| retval |
	retval := IdentityDictionary 
				newFrom: (self nodes collect: [:node | node -> (0 @ 0)]).
	self edges do: 
			[:edge | 
			| fromNode toNode delta mag force |
			fromNode := edge startConnection.
			toNode := edge endConnection.
			toNode notNil & fromNode notNil 
				ifTrue: 
					[delta := toNode center - fromNode center.
					delta := delta / scale.
					mag := delta r - 2.
					force := mag * delta.
					retval at: fromNode put: (retval at: fromNode) + force.
					retval at: toNode put: (retval at: toNode) - force]].
	^self nodes collect: [:ea | retval at: ea]! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 10/21/2003 16:42'!
applyForces
	| applied |
	applied := 0.
	self nodes with: self forces
		do: [:node :force | | newBounds |
			(force x > 1.0 or: [ force y > 1.0 ]) ifTrue: [
				newBounds := (node bounds translateBy: force) translatedToBeWithin: layoutMorph bounds.
				node position: newBounds origin asIntegerPoint.
				applied := applied + 1.]
		 ].
	^applied! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 10/21/2003 15:13'!
clearForces
	self forces: (Array new: self nodes size withAll: 0.0 @ 0.0)! !

!NCGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 10/21/2003 15:16'!
clearNodesAndConnections
	self edges: Array new.
	self nodes: Array new.
	self forces: Array new! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 10/21/2003 15:13'!
computeForces
	"Evaluate the weighted sum of my aesthetics, and apply the results to the nodes."

	self clearForces.
	aesthetics do: 
			[:assoc | 
			| weight gradient mag |
			gradient := (self perform: assoc key) asArray.
			weight := assoc value.
			self forces: (self forces with: gradient
						collect: [:prev :delta | prev + (delta * weight)])]! !

!NCGraphLayoutAlgorithm methodsFor: 'stepping' stamp: 'nk 8/16/2004 13:48'!
delete
	self stopStepping.! !

!NCGraphLayoutAlgorithm methodsFor: 'accessing' stamp: 'nk 10/21/2003 15:18'!
edges
	^Array withAll: (edges select: [ :ea | ea notNil ])! !

!NCGraphLayoutAlgorithm methodsFor: 'accessing' stamp: 'nk 10/21/2003 15:34'!
edges: anObject
	edges := WeakArray withAll: anObject asArray.! !

!NCGraphLayoutAlgorithm methodsFor: 'accessing' stamp: 'nk 10/21/2003 15:13'!
forces
	^forces! !

!NCGraphLayoutAlgorithm methodsFor: 'accessing' stamp: 'nk 10/21/2003 15:13'!
forces: anObject
	^forces := anObject! !

!NCGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 10/21/2003 15:44'!
initialize
	aesthetics := OrderedCollection new.
	self clearNodesAndConnections.
	self maxIterations: 2500.
	iterations := 0.
	self scale: 100.0! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'drm 7/27/2003 16:48'!
layout
	self setUpForLayout.
	maxIterations timesRepeat: [
		self step
	]! !

!NCGraphLayoutAlgorithm methodsFor: 'accessing' stamp: 'nk 10/21/2003 14:59'!
layoutMorph
	^layoutMorph! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 10/21/2003 16:40'!
layoutMorph: aPasteUpMorph 
	self nodes: aPasteUpMorph allConnectedMorphs.
	self edges: aPasteUpMorph allConnections.
	layoutMorph := aPasteUpMorph.
	aesthetics isEmpty ifTrue: [self setAesthetics].
	self clearForces.
	self setUpForLayout.
	self edges do: [ :e | e straighten ].
	iterations := 0.
	self startStepping! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'drm 7/10/2003 18:41'!
layoutWorld
	self layoutMorph: World! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 3/15/2002 17:41'!
limitForces
! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 3/15/2002 10:16'!
magnitudeOfVector: vector
	| mag |
	mag := vector inject: 0.0 into: [ :sum :pt | sum + (pt x * pt x) + (pt y * pt y) ].
	^mag sqrt! !

!NCGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 3/15/2002 11:47'!
maxIterations: max
	maxIterations := max.! !

!NCGraphLayoutAlgorithm methodsFor: 'accessing' stamp: 'nk 10/21/2003 15:18'!
nodes
	^Array withAll: (nodes select: [ :ea | ea notNil ])! !

!NCGraphLayoutAlgorithm methodsFor: 'accessing' stamp: 'nk 10/21/2003 15:32'!
nodes: anObject
	nodes := WeakArray withAll: anObject asArray! !

!NCGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 3/15/2002 15:55'!
scale: aNumber
	scale := aNumber! !

!NCGraphLayoutAlgorithm methodsFor: 'aesthetics' stamp: 'nk 10/21/2003 14:57'!
setAesthetics
	self edges size < (2 * self nodes size) 
		ifTrue: 
			[self addAesthetic: #aeCentripetal withWeight: 0.1.
			self addAesthetic: #aeEdgeFlowHang withWeight: 0.5.
			self addAesthetic: #aeMinEdgeLength withWeight: 0.2.
			self addAesthetic: #aeNodeRepulsion withWeight: 0.3.
			self addAesthetic: #aeEdgeRepulsion withWeight: 0.3.
			self addAesthetic: #aeSpringEdgeLength withWeight: 0.2]
		ifFalse: 
			[self addAesthetic: #aeCentripetal withWeight: 0.9.
			self addAesthetic: #aeEdgeFlowHang withWeight: 0.1.
			self addAesthetic: #aeMinEdgeLength withWeight: 0.1.
			self addAesthetic: #aeNodeRepulsion withWeight: 0.9.
			self addAesthetic: #aeEdgeRepulsion withWeight: 0.3.
			self addAesthetic: #aeSpringEdgeLength withWeight: 0.1]! !

!NCGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 10/21/2003 14:57'!
setScale
	"Iterate through the nodes and get a reasonable scale"

	self scale: (self nodes collect: [:ea | ea bounds extent]) max r! !

!NCGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'drm 7/27/2003 16:47'!
setUpForLayout
	self setScale! !

!NCGraphLayoutAlgorithm methodsFor: 'stepping' stamp: 'nk 10/21/2003 16:29'!
shouldGetStepsFrom: aWorld
	^true! !

!NCGraphLayoutAlgorithm methodsFor: 'stepping' stamp: 'nk 7/19/2004 00:40'!
startStepping
	iterations := iterations - (iterations \\ 3).
	layoutMorph ifNil: [^self].
	World
		startStepping: self
		at: Time millisecondClockValue
		selector: #step
		arguments: { 
				}
		stepTime: self stepTime! !

!NCGraphLayoutAlgorithm methodsFor: 'stepping' stamp: 'nk 10/21/2003 16:32'!
step
	"Evaluate the weighted sum of my aesthetics, and apply the results to 
	the nodes."

	(iterations \\ 3) caseOf: {
		[ 0 ] -> [ self computeForces ].
		[ 1 ] -> [ self limitForces ].
		[ 2 ] -> [ self applyForces ] }.

	iterations := iterations + 1.
	(iterations >= (maxIterations * 3)) ifTrue: [self stopStepping]! !

!NCGraphLayoutAlgorithm methodsFor: 'stepping' stamp: 'nk 10/21/2003 16:33'!
stepTime
	^50! !

!NCGraphLayoutAlgorithm methodsFor: 'stepping' stamp: 'nk 7/19/2004 00:40'!
stopStepping
	World stopStepping: self! !

!NCGraphLayoutAlgorithm methodsFor: 'stepping' stamp: 'nk 5/2/2004 22:21'!
world
	^layoutMorph ifNotNil: [ layoutMorph world ]! !

NCGraphLayoutAlgorithm subclass: #NCSimulatedAnnealingGraphLayoutAlgorithm
	instanceVariableNames: 'temperature initialTemperature endTemperature lambda'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ConnectorsGraphLayout'!

!NCSimulatedAnnealingGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 3/15/2002 11:26'!
endTemperature: aNumber
	endTemperature := aNumber.
! !

!NCSimulatedAnnealingGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 3/15/2002 11:26'!
initialTemperature: aNumber
	initialTemperature := aNumber.
! !

!NCSimulatedAnnealingGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 3/15/2002 11:26'!
initialize
	super initialize.
	initialTemperature := 100.0 .
	endTemperature := 0.001 .
! !

!NCSimulatedAnnealingGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 10/21/2003 15:13'!
limitForces
	| mag |
	mag := self magnitudeOfVector: self forces.
	mag > temperature 
		ifTrue: 
			[mag := temperature / mag.
			self forces: self forces * mag]! !

!NCSimulatedAnnealingGraphLayoutAlgorithm methodsFor: 'initialization' stamp: 'nk 3/15/2002 11:25'!
setLambda
	lambda := (initialTemperature / endTemperature) raisedTo: (1.0 / maxIterations).
! !

!NCSimulatedAnnealingGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'drm 7/16/2003 18:10'!
setUpForLayout
	super setUpForLayout.   "super setUpForLayout sets the scale, which avoids bounds errors"
	self setLambda.
	temperature := initialTemperature.
! !

!NCSimulatedAnnealingGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 3/15/2002 17:58'!
step
	super step.
	temperature := temperature / lambda.
! !

!NCSimulatedAnnealingGraphLayoutAlgorithm methodsFor: 'layout' stamp: 'nk 8/16/2004 13:52'!
temperature
	^temperature! !
